--[=====[
		## RM Utils library ver. 1.0.1
		## rmUtils.lua - code module
		Code for rmUtils lib
--]=====]

local RMUTILS10 = "rmUtils-1.0"
local RMUTILS10_MINOR = 2 -- Bump on changes

if not LibStub then
	error(RMUTILS10 .. " requires LibStub.")
end

local U = LibStub:NewLibrary(RMUTILS10, RMUTILS10_MINOR)
if not U then return end

local ipairs = ipairs
local pairs = pairs
local print = print
local rawget = rawget
local strrep = string.rep
local strformat = string.format
local select = select
local setmetatable = setmetatable
local tinsert = table.insert
local tremove = table.remove
local tonumber = tonumber
local tostring = tostring
local type = type
local unpack = unpack
local math_floor = math.floor

local BreakUpLargeNumbers = BreakUpLargeNumbers
local CreateColor = CreateColor
local GetBuildInfo = GetBuildInfo
local GetLocale = GetLocale
local LARGE_NUMBER_SEPERATOR = LARGE_NUMBER_SEPERATOR
local BLACK_FONT_COLOR = BLACK_FONT_COLOR
local RED_FONT_COLOR = RED_FONT_COLOR
local ORANGE_FONT_COLOR = ORANGE_FONT_COLOR
local LIGHTBLUE_FONT_COLOR = LIGHTBLUE_FONT_COLOR
local GREEN_FONT_COLOR = GREEN_FONT_COLOR
local NORMAL_FONT_COLOR = NORMAL_FONT_COLOR
local WOW_PROJECT_ID = WOW_PROJECT_ID
local WOW_PROJECT_CLASSIC = WOW_PROJECT_CLASSIC

-- Remove all known globals after this point
-- luacheck: std none

--[[ Classic WoW detection ]]

if WOW_PROJECT_ID and WOW_PROJECT_CLASSIC then
	U.IsWoWClassic = WOW_PROJECT_ID == WOW_PROJECT_CLASSIC
else
	U.IsWoWClassic = select(4, GetBuildInfo()) < 20000
end

--[[ Empty Function ]]
U.EmptyFn = function() end

--[[ Debug Localization ]]
do
	U.DebugL = function(L)
		return setmetatable({}, { __index = function(self, key) return L and rawget(L, key) or key end })
	end
end

--[[ Large number formatting ]]
local breakUpLargeNumber
do
	local locale = GetLocale()
	local breakCount = 3

	local thousandSeparator = {
		deDE = ".",
		frFR = "\194\160", -- NBSP
		esES = ".",
		ptBR = ".",
		ruRU = "\194\160", -- NBSP
	}

	local classSeparator = thousandSeparator[locale]

	if not classSeparator and not U.IsWoWClassic then
		breakUpLargeNumber = BreakUpLargeNumbers
	else
		if not classSeparator then
			classSeparator = LARGE_NUMBER_SEPERATOR and LARGE_NUMBER_SEPERATOR:len() > 0
								and LARGE_NUMBER_SEPERATOR
								or ","
		end
		breakUpLargeNumber = function(num)
			if type(num) ~= "number" then
				return num
			end

			local strnum = tostring(num)
			local _, _, minus, whole, _, fraction = strnum:find("^(%-?)(%d+)(%.?)(%d*)")

			if #whole <= breakCount then
				return strnum
			end

			local fullBreaks = math_floor(#whole / breakCount)
			local remain = #whole % breakCount
			local broken = ""

			for i = 0, fullBreaks - 1 do
				if #broken > 0 then
					broken = classSeparator .. broken
				end
				local ind1, ind2 = - (i * 3 + 3), - (i * 3 + 1)
				broken = whole:sub(ind1, ind2) .. broken
			end

			if remain > 0 then
				broken = whole:sub(1, remain) .. classSeparator .. broken
			end

			-- add localized 'dot' to fraction digits
			fraction = strformat("%0." .. #fraction .. "F", tonumber("0." .. fraction)):sub(2)

			return minus .. broken .. fraction
		end
	end

	function U.Commify(num)
		if type(num) ~= "number" then
			return num
		end
		return breakUpLargeNumber(num)
	end
end

--[[ Message formatting ]]

do
	local function ts(icon, color)
		return { icon = icon, color = color }
	end
	local iconFormat = "|T%s:0|t\32"
	local textSettings = {
		fatal = ts([[interface\targetingframe\ui-raidtargetingicon_8]], BLACK_FONT_COLOR),
		error = ts([[interface\targetingframe\ui-raidtargetingicon_7]], RED_FONT_COLOR),
		warning = ts([[interface\targetingframe\ui-raidtargetingicon_2]], ORANGE_FONT_COLOR),
		info = ts([[interface\targetingframe\ui-raidtargetingicon_5]], LIGHTBLUE_FONT_COLOR),
		instruction = ts([[interface\targetingframe\ui-raidtargetingicon_4]], GREEN_FONT_COLOR),
		key = ts(nil, CreateColor(1, 1, 0.125, 1)),
		normal = ts(nil, NORMAL_FONT_COLOR),
	}
	local function getIndex(_, index)
		if type(index) == "string" then
			local key = index:match("^Get(%w+)"):lower()
			local textSetting = textSettings[key]
			if not textSetting.func then
				local icon, color = textSetting.icon, textSetting.color
				textSetting.func = function(text)
					local iconTag = icon and iconFormat:format(icon) or ""
					return iconTag .. color:WrapTextInColorCode(text)
				end
			end
			return textSetting.func
		end
	end
	U.Text = setmetatable({}, { __index = getIndex })
end

--[[ Function functions ]]

function U.Bind(func, ...)
	local presetArgs = { ... }
	return function(...)
		func(unpack(U.Append(presetArgs, { ... })))
	end
end

--[[ Table functions ]]

do
	local u_override, u_merge, u_clone, u_allValues, u_namedValues

	function u_override(target, source)
		if not source then
			target, source = {}, target
		elseif type(target) ~= "table" then
			target = {}
		end
		for k, v in pairs(source) do
			if type(v) == "table" then
				target[k] = u_override(target[k], v)
			else
				target[k] = v
			end
		end
		return target
	end

	function u_merge(target, source)
		if not source then
			target, source = {}, target
		elseif type(target) ~= "table" then
			target = {}
		end
		for k, v in pairs(source) do
			if target[k] == nil then
				if type(v) == "table" then
					target[k] = u_merge(target[k], v)
				else
					target[k] = v
				end
			end
		end
		return target
	end

	function u_clone(source)
		return u_merge(source)
	end

	function u_allValues(t)
		local res = {}
		for _, v in pairs(t) do
			tinsert(res, v)
		end
		return res
	end

	function u_namedValues(t, names)
		local res = {}
		for _, v in ipairs(names) do
			tinsert(res, t[v])
		end
		return res
	end

	function U.ForEach(t, forEachFunc)
		for k, v in pairs(t) do
			forEachFunc(v, k, t)
		end
	end

	function U.Filter(t, filterFunc)
		local res = {}
		for k, v in pairs(t) do
			if filterFunc(v, k, t) then
				if type(k) == "number" then
					tinsert(res, v)
				else
					res[k] = v
				end
			end
		end
		return res
	end

	function U.Map(t, mapFunc)
		local res = {}
		for k, v in pairs(t) do
			res[k] = mapFunc(v, k, t)
		end
		return res
	end

	function U.Append(...)
		local result, num, tbl = {}, 1, {...}
		for i = 1, #tbl do
			local t = tbl[i]
			if t then
				for j = 1, #t do
					result[num], num = t[j], num + 1
				end
			end
		end
		return result
	end

	function U.AppendTable(t, ...)
		local result, num, tbl = t or {}, 1, {...}
		for i = 1, #tbl do
			local tt = tbl[i]
			if tt then
				for j = 1, #tt do
					result[num], num = tt[j], num + 1
				end
			end
		end
		return result
	end

	function U.Slice(t, start, finish)
		local result = {}
		for i = start, finish do
			result[#result + 1] = t[i]
		end
		return result
	end

	function U.SliceLength(t, start, length)
		return U.Slice(t, start, start + length)
	end

	function U.Push(t, item)
		tinsert(t, item)
	end

	function U.Pop(t)
		return tremove(t)
	end

	function U.Shift(t)
		return tremove(t, 1)
	end

	function U.Unshift(t, item)
		tinsert(t, 1, item)
	end

	function U.Keys(t)
		local res = {}
		for k, _ in pairs(t) do
			tinsert(res, k)
		end
		return res
	end

	function U.Values(t, ...)
		local names = ...
		if not names then
			return u_allValues(t)
		elseif type(names) == "table" then
			return u_namedValues(t, names)
		else
			return u_namedValues(t, {...})
		end
	end

	U.Override = u_override
	U.Merge = u_merge
	U.Clone = u_clone
end

--[[ Text formatting and replacing ]]

function U.FormatPercents(dividend, divisor)
	dividend = dividend or 0
	divisor = divisor or 0
	if divisor == 0 then
		return 0
	end
	return ("%.1F%%%%"):format(dividend / divisor * 100)
end

function U.FormatRemainingPercents(dividend, divisor)
	dividend = dividend or 0
	divisor = divisor or 0
	if divisor == 0 then
		return 0
	end
	return ("%.1F%%%%"):format(100 - (dividend / divisor * 100))
end

function U.MultiReplace(text, replacements)
	if not text or text == "" then
		return ""
	end
	local result = text
	U.ForEach(
		replacements,
		function(value, pattern)
			result = result:gsub(pattern, value)
		end
	)
	return result
end

--[[ Debug functions ]]

function U.PrintThru(value, label)
	--@debug@
	local text = tostring(value)
	if label then
		text = label .. ": " .. text
	end
	print(text)
	--@end-debug@
	return value
end

function U.PrintTypes(tbl, title)
	if title and #title > 0 then
		print(title)
	end
	if type(tbl) == "table" then
		for k, v in pairs(tbl) do
			print(k, "(" .. type(v) .. "):", v)
		end
	else
		print("value (" .. type(tbl) .. "):", tbl)
	end
end

function U.PrintTable(table, tableName, maxLevel)
	--@debug@
	local function indent(i)
		return strrep("  ", i)
	end
	local function printIndent(i, s)
		print(indent(i) .. s)
	end
	local function printTableImpl(t, name, level, max)
		if level > max then
			return
		elseif t == nil then
			-- this is possible only for top level value
			printIndent(level, name .. " = nil,")
			return
		end

		printIndent(level, name .. " = {")

		for k, v in pairs(t) do
			local vType = type(v)
			if vType == "table" then
				printTableImpl(v, k, level + 1, max)
			elseif vType == "string" then
				printIndent(level + 1, k .. " = '" .. tostring(v) .. "',")
			else
				printIndent(level + 1, k .. " = " .. tostring(v) .. ",")
			end
		end
		printIndent(level, "},")
	end

	printTableImpl(table, tableName or "<unnamed>", 0, maxLevel or 16)
	--@end-debug@
end
